/*
 * Copyright (C) 2002 RealVNC Ltd.  All Rights Reserved.
 *
 * This is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this software; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307,
 * USA.
 */

#ifndef __ZRLE_TYPES_H__
#define __ZRLE_TYPES_H__

typedef unsigned char  zrle_U8;
typedef unsigned short zrle_U16;
typedef unsigned int   zrle_U32;
typedef signed char    zrle_S8;
typedef signed short   zrle_S16;
typedef signed int     zrle_S32;

#endif /* __ZRLE_TYPES_H__ */
